#include <cstdlib>
#include <fstream>
#include <iostream>

#include <misc/utils.h>


using namespace CSA;


std::string gapChars = " =-_";
std::string bases = "ACGT";

usint
cleanAlignment(std::ifstream& input, std::ofstream& output, usint sequences, bool quality_included)
{
  input.clear();
  input.seekg(0, std::ios_base::beg);

  usint lines = 0;
  usint chars[CHARS];
  for(usint c = 0; c < CHARS; c++) { chars[c] = 0; }

  std::string line;
  if(quality_included) { getline(input, line); }
  usint multi = (quality_included ? 2 : 1);
  while(!input.eof())
  {
    getline(input, line);
    if(line.empty()) { continue; }

    for(usint i = 0; i < sequences; i++)
    {
      char t = toupper(line[i * multi]);
      if(gapChars.find(t) != std::string::npos) { t = ' '; }
      chars[(int)t]++;
      output.put(t);
    }
    output.put('\n');
    lines++;
  }

  std::cout << "Character counts:" << std::endl;
  for(usint c = 0; c < CHARS; c++)
  {
    if(chars[c] > 0) { std::cout << "  " << (char)c << " (" << c << "): " << chars[c] << std::endl; }
  }
  std::cout << std::endl;

  return lines;
}


void
percolateDown(usint i, usint j, char** content, usint* gap_start, usint* gap_stop)
{
  content[i][j] = content[gap_start[j] - 1][j];
  content[gap_start[j] - 1][j] = ' ';
  while(gap_start[j] > 0 && content[gap_start[j] - 1][j] == ' ') { gap_start[j]--; }
  gap_stop[j]--;
}

usint
getMostCommon(std::string& allowed, usint* counts)
{
  usint most_common = bases.length();

  for(usint i = 0; i < allowed.length(); i++)
  {
    usint pos = bases.find(allowed[i]);
    if(pos == std::string::npos) { continue; }

    if(most_common < bases.length())
    {
      if(counts[pos] > counts[most_common]) { most_common = pos; }
    }
    else if(counts[pos] > 0) { most_common = pos; }
  }

  return most_common;
}

// Handle IUPAC codes.
// We replace symbols allowing multiple bases with the most common of them at the current position.
// We create a gap if no allowed base is found at the current position.
usint
replaceByMostCommonBase(char* line, usint sequences)
{
  usint counts[4]; for(usint i = 0; i < 4; i++) { counts[i] = 0; }
  for(usint i = 0; i < sequences; i++)
  {
    std::size_t temp = bases.find(line[i]);
    if(temp != std::string::npos) { counts[temp]++; }
  }

  usint replaced = 0;
  for(usint i = 0; i < sequences; i++)
  {
    std::string allowed = "";
    switch(line[i])
    {
      case 'R':
        allowed = "AG"; break;
      case 'Y':
        allowed = "CT"; break;
      case 'K':
        allowed = "GT"; break;
      case 'M':
        allowed = "AC"; break;
      case 'S':
        allowed = "CG"; break;
      case 'W':
        allowed = "AT"; break;
      case 'B':
        allowed = "CGT"; break;
      case 'D':
        allowed = "AGT"; break;
      case 'H':
        allowed = "ACT"; break;
      case 'V':
        allowed = "ACG"; break;
      case 'N':
        allowed = "ACGT"; break;
    }
    if(allowed.length() == 0) { continue; }

    usint most_common = getMostCommon(allowed, counts);
    if(most_common >= bases.length()) { line[i] = ' '; }
    else { line[i] = bases[most_common]; }
    replaced++;
  }

  return replaced;
}


// Each gap must be preceded a character that does not occur at the last line of the gap.
// Additionally, two gaps ending at the same line but beginning at different lines must be
// preceded by different characters.
usint
handleGaps(char* filename, usint sequences, usint lines)
{
  std::ifstream input(filename, std::ios_base::binary);
  char** content = new char*[lines];
  bool* ok = new bool[lines];
  usint replaced = 0;
  for(usint i = 0; i < lines; i++)
  {
    content[i] = new char[sequences + 1];
    input.read(content[i], sequences + 1);
    replaced += replaceByMostCommonBase(content[i], sequences);
    ok[i] = false;
  }
  input.close();

  usint gap_start[sequences];
  usint gap_stop[sequences];
  for(usint i = 0; i < sequences; i++) { gap_start[i] = lines; gap_stop[i] = lines; }

  std::ofstream output(filename, std::ios_base::binary);
  usint moved = 0, deleted = 0;
  for(usint l = lines; l > 0; l--)
  {
    usint i = l - 1;

    for(usint j = 0; j < sequences; j++)
    {
      if(content[i][j] == ' ')
      {
        if(gap_start[j] > i)  // New gap.
        {
          gap_start[j] = i;
          while(gap_start[j] > 0 && content[gap_start[j] - 1][j] == ' ') { gap_start[j]--; }
          gap_stop[j] = i;
        }
      }
      else
      {
        gap_start[j] = gap_stop[j] = i + 1;
      }
    }

    // Determine the characters occurring at line i.
    usint counts[4]; for(usint j = 0; j < 4; j++) { counts[j] = 0; }
    for(usint j = 0; j < sequences; j++)
    {
      std::size_t temp = bases.find(content[i][j]);
      if(temp != std::string::npos) { counts[temp]++; ok[i] = true; }
    }

    if(ok[i]) // There is a character at line i.
    {
      // If two gaps beginning at different lines end at line i, and are preceded by
      // the same characters, move these characters to line i.
      for(usint j = 0; j < sequences; j++)
      {
        usint mov = 0;
        if(gap_stop[j] != i) { continue; }  // Not a gap end.
        for(usint jj = j + 1; jj < sequences; jj++)
        {
          if(gap_stop[jj] != i) { continue; } // Not a gap end.
          if(gap_start[j] == gap_start[jj]) { continue; } // Not a problem.
          if(gap_start[j] == 0 || gap_start[jj] == 0) { continue; } // Not a problem.
          if(content[gap_start[j] - 1][j] == content[gap_start[jj] - 1][jj])
          {
            percolateDown(i, jj, content, gap_start, gap_stop); mov++;
            std::size_t temp = bases.find(content[i][jj]);
            if(temp != std::string::npos) { counts[temp]++; ok[i] = true; }
          }
        }
        if(mov > 0)
        {
          percolateDown(i, j, content, gap_start, gap_stop); mov++;
          moved += mov;
          std::size_t temp = bases.find(content[i][j]);
          if(temp != std::string::npos) { counts[temp]++; ok[i] = true; }
        }
      }

      // Move characters preceding a gap to the end of the gap if they already occur there.
      for(usint j = 0; j < sequences; j++)
      {
        if(gap_stop[j] == i && gap_start[j] > 0)
        {
          std::size_t temp = bases.find(content[gap_start[j] - 1][j]);
          if(temp != std::string::npos && counts[temp] > 0)
          {
            percolateDown(i, j, content, gap_start, gap_stop);
            counts[temp]++; moved++;
          }
        }
      }
    }
    else  // Gaps ending at line i will now end at line i - 1.
    {
      for(usint j = 0; j < sequences; j++)
      {
        if(gap_stop[j] == i) { gap_stop[j]--; }
      }
      deleted++;
    }
  }

  usint truelines = 0;
  for(usint i = 0; i < lines; i++)
  {
    if(ok[i]) { output.write(content[i], sequences + 1); truelines++; }
  }
  output.close();

  std::cout << "Lines (final): " << truelines << std::endl;
  std::cout << "Moved " << moved << " characters." << std::endl;
  std::cout << "Replaced " << replaced << " characters." << std::endl;
  std::cout << "Deleted " << deleted << " lines." << std::endl;
  std::cout << std::endl;

  for(usint i = 0; i < lines; i++) { delete[] content[i]; }
  delete[] content;
  delete[] ok;

  return truelines;
}


int
main(int argc, char** argv)
{
  std::cout << "Alignment cleaner" << std::endl;
  std::cout << std::endl;
  if(argc < 3)
  {
    std::cout << "Usage: clean_alignment alignment_file output_file" << std::endl;
    return 1;
  }

  std::cout << "Alignment file: " << argv[1] << std::endl;
  std::ifstream alignment_file(argv[1], std::ios_base::binary);
  if(!alignment_file)
  {
    std::cerr << "Error opening alignment file!" << std::endl;
    return 2;
  }

  std::cout << "Output file: " << argv[2] << std::endl;
  std::ofstream output_file(argv[2], std::ios_base::binary);
  if(!output_file)
  {
    std::cerr << "Error creating output file!" << std::endl;
    return 3;
  }

  // Determine the format and the number of sequences.
  std::string line;
  bool quality_included = false;
  std::getline(alignment_file, line);
  usint sequences = line.size();
  if(line[0] == '>')
  {
    quality_included = true; 
    std::getline(alignment_file, line);
    sequences = line.size() / 2;
  }
  std::cout << "Sequences: " << sequences << std::endl;
  std::cout << std::endl;

  usint lines = cleanAlignment(alignment_file, output_file, sequences, quality_included);
  output_file.close();
  alignment_file.close();
  std::cout << "Lines (initial): " << lines << std::endl;

  handleGaps(argv[2], sequences, lines);

  return 0;
}
