#ifndef GRAPH_H
#define GRAPH_H


#include <fstream>
#include <iostream>
#include <vector>


#include <misc/definitions.h>


namespace CSA
{

//--------------------------------------------------------------------------

struct PathNodeList;

// Indegree must be set to 0 after use before the node is deleted.
struct GraphNode
{
  usint label;
  usint value;  // Used for sampling in GCSA.
  union
  {
    PathNodeList* paths;
    usint         indegree;
  };
  usint outdegree;  // This often doubles as node id.
};

struct GraphEdge
{
  GraphNode* from;
  GraphNode* to;
};

//--------------------------------------------------------------------------

struct PathNode
{
  usint first;
  usint second;

  bool sorted;

  GraphNode* from;
  GraphNode* to;

  PathNode* left;
  PathNode* right;

  PathNode();
  ~PathNode();
};

struct PathEdge
{
  PathNode* from;
  PathNode* to;
  PathNode* extra;

  PathEdge();
  PathEdge(PathNode* _from, PathNode* _to, PathNode* _extra = 0);
};

struct PathNodeList
{
  PathNode*     item;
  PathNodeList* next;

  PathNodeList(PathNode* _item, PathNodeList* _next);

  bool contains(PathNode* _item);
};

void deleteList(PathNodeList* list);

//--------------------------------------------------------------------------

class PathGraph;

class Graph
{
  public:
    Graph(const std::string& _node_name, const std::string& _edge_name, bool load);
    ~Graph();

    void deletePaths();
    void deleteNodes();
    void deleteEdges();

    void write();

    // Reorders the nodes to allow better sampling.
    // Do not use if there is an PathGraph attached to the Graph.
    void reorder();

    void sortEdges();
    pair_type getEdgesFrom(GraphNode* node);  // Use sortEdges() first.

    void checkGraph();

    // Writes node ids to node->outdegree.
    // Sorting nodes invalidates the ids.
    void setIds();

    PathGraph* getLastGeneration(bool minimal = false); // See PathGraph constructor.

    GraphNode** nodes;
    GraphEdge*  edges;
    usint node_count, edge_count;
    const std::string node_name, edge_name;
};

//--------------------------------------------------------------------------

class PathGraph
{
  public:
    // load:    load the PathGraph from disk
    // minimal: do not create paths lists on GraphNodes
    PathGraph(Graph& _parent, usint _generation, bool load, bool minimal = false);
    ~PathGraph();

    void deleteNodes();
    void deleteEdges();

    std::string getNodeName();
    std::string getEdgeName();

    bool exists();
    void load(bool minimal = false);  // See constructor.
    void write();

    void sortEdges(bool by_destination = false);
    pair_type getEdgesFrom(PathNode* from);
    pair_type getEdgesTo(PathNode* to);

    // We assume that nodes contain their ids as pn->second.
    void checkGraph();
    void printNodeInfo(usint i);

    // Returns SA samples as (index, value). Only works on a sorted graph.
    // User must delete the samples.
    std::vector<pair_type>* getSamples(usint sample_rate);

    PathNode** nodes;
    PathEdge* edges;
    usint node_count, edge_count;

    Graph& parent;
    usint generation;
};

//--------------------------------------------------------------------------

// We always assume that the parameters are valid.
// Some parts of graph may be deleted in the function.

// Build paths of length up to 2.
bool initPaths(Graph& graph);

// Combine paths of length n to make paths of length 2n.
bool doublePaths(Graph& graph);


//--------------------------------------------------------------------------

/*
  Pseudocode versions of the algorithms.


  // Builds the index. OK
  function buildIndex():
    sorted = initPaths()
    while not sorted:
      sorted = doublePaths()
    getBWT()


  // Paths of length up to 2. OK. Implemented.
  function initPaths():
    for each GraphNode gn:
      for each GraphEdge (gn, suc):
        pn = new PathNode(gn.label, suc.label, gn, suc, NIL, NIL)
        gn.paths.append(pn)

    // to == NIL means that the path is already sorted
    new PathNode(lastNode.label, 0, lastNode, NIL, NIL, NIL)

    for each PathNode pn:
      for each suc in pn.to.paths:
        new PathEdge(pn, suc)

    return updateLabels()


  // Updates the labels after doubling. OK. Implemented.
  function updateLabels():
    prev = NIL
    rank = 0
    start = NIL

    sort PathNodes

    for each PathNode pn in lexicographic order:
      if (pn.first, pn.second) != prev:
        rank = rank + 1
        prev = (pn.first, pn.second)
        if start != NIL:
          start.to = NIL  // already sorted
        start = pn
      else:
        sorted = false
        start = NIL
      (pn.first, pn.second) = (rank, 0)

    if start != NIL:
      start.to = NIL  // already sorted
    return sorted


  // Doubles the length of paths from n to 2n. OK
  function doublePaths():

    // Part 1: Determine the new PathNodes. OK
    // There are some omissions. See the actual code.
    for each PathNode pn:
      if pn.to == NIL:  // already unique
        new PathNode(pn.first, 0, pn.from, NIL, pn, NIL)
      else:
        for each (pn.to, suc) in GraphEdges:
          for each path in suc.paths:
            temp = new PathNode(pn.first, path.first, pn.from, path.to, pn, path)
            pn.left.append(temp)
            path.right.append(temp)

    // Part 2: Determine the new PathEdges. OK
    //   left = left(successors(left(pn)))
    //   right = right(successors(right(pn)))
    //   successors(pn) = intersect(left, right)
    for each new PathNode pn:
      successors = NIL
      for each (pn.left.item, suc) in old PathEdges:
        successors.append(suc.left)
      if pn.to != NIL:
        right = NIL
        for each (pn.right.item, suc) in old PathEdges:
          right.append(suc.right)
        successors = intersect(successors, right)
      for each distinct suc in successors:
        new PathEdge(pn, suc)

    // Part 3: Get rid of the old path graph. OK
    delete old PathNodes
    delete old PathEdges
    for each GraphNode gn:
      gn.paths = gn.newpaths
      gn.newpaths = NIL

    // Part 4: Sort the new PathNodes and update their labels. OK
    updateLabels()

    // Part 5: Merge PathNodes when possible. OK
    // A maximal set is either one PathNode or a set of adjacent ranks (PathNode.first values) such that
    // the set contains only one PathNode.from value.
    for each maximal set S of PathNodes:
      pn = new PathNode(S.item.value, 0, S.item.first, NIL, NIL, NIL)
      from = NIL
      to = NIL
      for each node in S:
        for each (pred, node) in PathEdges:
          from.append(pred)
          delete (pred, node)
        for each (node, suc) in PathEdges:
          to.append(suc)
          delete (suc, node)
        delete node
      for each distinct pred in from:
        new PathEdge(from, pn)
      for each distinct suc in to:
        new PathEdge(pn, to)

    // Part 6: Check sortedness. OK
    prev = NIL
    for each PathNode pn in lexicographic order:
      if pn.first == prev:
        return false
      prev = pn.first
    return true


  // Reads (BWT, firstChildren, hasPredecessor) from the sorted paths. OK
  // hasPredecessor[i] = 1 if BWT[i] != NIL (a merge happens here)
  function getBWT():
    for each PathNode pn in lexicographic order:
      n = sum of (pn, suc) in PathEdges   // We are merging n paths.
      first = true
      for each (pred, pn) in PathEdges:   // Node pred is sorted by path pn.
        if first:
          firstChildren.append(1)
          firstChildren.append(0, n - 1)
          first = false
        else:
          firstChildren.append(0, n)
        BWT.append(pred.from.label)
        BWT.append(NIL, n - 1)
        hasPredecessor.append(1)
        hasPredecessor.append(0, n - 1)
*/


} // namespace CSA


#endif // GRAPH_H

